/*
=============================================================
proxy.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the invisible proxy program. The invisible proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover sleeping devices and make sure they still seem to be connected
to the network by spoofing HTTP OK and SSDP:alive. When there is
an incoming TCP connection for the device the proxy will wake it up using
WOL.

This file contains the main function for the program. It will start 
by making a discovery of the complete network. Then the proxy sniffs packets
and process them. ARP and discovery for sleeping devices are answered.
SSDP:alive messages are added to the data structure. Incoming TCP
connections to sleeping devices result in a WOL to the sleeping device.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netwib.h>
#include "proxy_funcs.h"



int main(int argc, char *argv[])
{
//--------------declaration-------------------------------
	netwib_buf msearchbuf, notifybuf, httpokbuf, device, rest, tempbuf, cmpbuf, filter;
	netwib_iphdr iphdr;
	netwib_udphdr udphdr;
	netwib_etherhdr ethhdr;
	netwib_arphdr arphdr;
	netwib_cmp cmp;
	netwib_err ret;	
	netwib_ip myip, broadcastip;
	netwib_port myport;
	netwib_time t, t2;
	netwib_bool b, iptest;
	netwib_uint16 tmp;
	netwib_io *pio, *piostorage, *pio2;
	netwib_ptr ptr;
	upnp_device *upnpdev;
	service *upnp_service;
	char str[100];
	netwib_ring_index *pringindex, *pringindex2;
	netwib_ring* upnp_devices;
    netwib_ring* proxy_devices;
    netwib_ring* thread_ring;
    netwib_ips *ips;
	/////////////////////////////////////////////////////////
	netwib_thread *cache_check; // thread to check and update cache
	netwib_thread *notify_check; // thread to send notify if needed
	netwib_bool event;
	netwib_ptr thread_return;
	////////////////////////////////////////////////////////
	
//--------------end of declaration------------------------
	
	netwib_init();
	if(argc != 4){
            netwib_fmt_display("Usage proxy [ip address of proxy] [port of proxy] [minutes before start proxying]\n");
            netwib_fmt_display("Example proxy 1.2.3.4 5678 2");
            exit(EXIT_FAILURE);
            }
	b = NETWIB_FALSE;
	//create and initiate buffers for masking of packets
	netwib_er(netwib_buf_init_mallocdefault(&device));
	netwib_er(netwib_buf_init_mallocdefault(&rest));
	netwib_er(netwib_buf_init_mallocdefault(&msearchbuf));
	netwib_er(netwib_buf_init_mallocdefault(&httpokbuf));
	netwib_er(netwib_buf_init_mallocdefault(&notifybuf));
	netwib_er(netwib_buf_init_mallocdefault(&cmpbuf));
	netwib_buf_init_ext_text("M-SEARCH", &tempbuf);//start of http when it is a m-search
	netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &msearchbuf);
	netwib_buf_init_ext_text("NOTIFY", &tempbuf); // start of hht when it is a sspd-notify
	netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &notifybuf);
	netwib_buf_init_ext_text("HTTP/1.1 200 OK", &tempbuf);//start of http when it is a m-search
	netwib_buf_decode(&tempbuf, NETWIB_DECODETYPE_DATA, &httpokbuf);
	netwib_ips_initdefault(&ips);
	netwib_er(netwib_ip_init_ip4_fields(239, 255,255,250, &broadcastip));
	netwib_ips_add_ip(ips, &broadcastip);
	netwib_buf_init_ext_text(argv[1], &tempbuf);
	netwib_ip_init_buf(&tempbuf, NETWIB_IP_DECODETYPE_IP4, &myip);
    myport = atoi(argv[2]);
    min_bef_proxy = atoi(argv[3]);
    //initiate the caches
    netwib_er(netwib_ring_init(&delete_device_from_list, NULL, &upnp_devices));
    netwib_er(netwib_ring_init(&delete_device_from_list, NULL, &proxy_devices));
    netwib_er(netwib_ring_init(NULL, NULL, &thread_ring)); // threadring will contain proxy_devices followed by upnp_devices
    ptr = (netwib_ptr *)ips;
    netwib_er(netwib_ring_add_first(thread_ring, ptr));
	ptr = (netwib_ptr *)upnp_devices;
    netwib_er(netwib_ring_add_first(thread_ring, ptr));
    ptr = (netwib_ptr *)proxy_devices;
    netwib_er(netwib_ring_add_first(thread_ring, ptr));
    
    

	ret = init_device(&device, NETWIB_DEVICE_HWTYPE_ETHER); //initiate an ethernet NIC
	if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error when initiating device\n"));
	}
	
	//send out an M-search
	ret = send_discovery(myport);
	if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error when sending m-search\n"));
	}
	netwib_er(netwib_fmt_display("M-search sent\n"));
	
	//initiate filter and time filter should be
	//udp and dst host [localip] and dst port [localport]
    netwib_er(netwib_buf_init_mallocdefault(&filter));
    netwib_er(netwib_buf_append_text("udp and dst host ", &filter));
    netwib_er(netwib_buf_append_ip(&myip,NETWIB_IP_ENCODETYPE_IP, &filter)); 
    netwib_er(netwib_buf_append_text(" and dst port ", &filter));
    netwib_er(netwib_buf_append_port(myport, &filter));
    netwib_fmt_display("Filter: %{buf}\n", &filter);
    netwib_er(netwib_time_init_now(&t));
    netwib_er(netwib_time_plus_sec(&t, 30));//add 30 seconds to actual time 
        
    ret =netwib_io_init_sniff(&device, &filter, &pio); // initiate sniff session
	if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error while initiating sniff session."));
	}
	//open dummy socket to bind the port
	 ret = netwib_io_init_sock_udp_ser(&myip, myport, &pio2);
   if(ret != NETWIB_ERR_OK)
   {
          netwib_er(netwib_fmt_display("Error when opening socket in main.\n"));
   }

	
    //sniff for HTTP OK for 30 sec
    //every sniff session is at least 2 sec
	while(!b){
             netwib_er(netwib__buf_reinit(&rest));
            netwib_er(netwib_time_init_now(&t2));
            netwib_er(netwib_time_plus_sec(&t2, 2));//add 2 seconds to actual time
	         //receive all HTTP OK
	         ret = sniff(&iphdr, &ethhdr, &udphdr, &rest,&t2, pio);	         
	         if(ret == NETWIB_ERR_OK){
                    //check if we really have a HTTP OK
                    netwib__buf_erase(&cmpbuf);
                    netwib_buf_append_buf(&rest, &cmpbuf);
	                cmpbuf.endoffset =cmpbuf.beginoffset+ 15;
	                netwib_buf_cmp(&cmpbuf, &httpokbuf, &cmp);
	                
		            if(cmp == NETWIB_CMP_EQ) {//received a HTTP OK
                           netwib_fmt_display("Received HTTP OK\n");	
                          process_packet(&iphdr, &ethhdr, &udphdr, &rest, HTTP_OK, upnp_devices);                          
                   } //end of CMP      	
    	         	                   	
                }
            netwib_er(netwib_time_iselapsed(&t, &b)); 	
            
    } //end of while
    netwib_er(netwib_io_close(&pio));
    
    /////////////////////////////////////////////////////////////////////////////
	///////////          initiate cache thread             //////////////////////
	netwib_er(netwib_thread_init(&update_cache, thread_ring, &cache_check));
	netwib_er(netwib_thread_init(&spoof_notify, thread_ring, &notify_check));
	////////////////////////////////////////////////////////////////////////////
    
	netwib_er(netwib_fmt_display("HTTP OK processed\n"));
	netwib_io_close(&pio2);
	
	//netwib_er(netwib_time_init_time( NETWIB_TIME_INFINITE,&t)); // infinite time
	netwib_er(netwib_io_init_storage(&piostorage));
    
    netwib__buf_erase(&filter);
    netwib_er(netwib_buf_append_text("udp or tcp or arp", &filter));
	netwib_er(netwib_buf_init_mallocdefault(&cmpbuf));
	ret =netwib_io_init_sniff(&device, &filter, &pio); // initiate sniff session
	netwib_er(netwib_io_plug_read(piostorage, pio));
	if(ret != NETWIB_ERR_OK){
    		netwib_er(netwib_fmt_display("Error while initiating sniff session."));
    	} 
	while(NETWIB_TRUE){           
		netwib__buf_reinit(&rest);
		netwib__buf_reinit(&cmpbuf);
		netwib_er(netwib_time_init_now(&t2));
        netwib_er(netwib_time_plus_sec(&t2, 2));//add 2 seconds to actual time
		ret =sniff(&iphdr, &ethhdr, &udphdr, &rest,&t2, piostorage);//sniff packets and get the headers		
		//if ip id is 123 then this is our packet that we spoofed, ignore it!
		//netwib_fmt_display("Ip id is %{uint16}\n", iphdr.header.ip4.id );
		if (ret == NETWIB_ERR_OK && iphdr.header.ip4.id != 123) {
		  netwib_ips_contains_ip(ips, &iphdr.dst, &iptest); // check if we should take care of this message
		  if(iptest) {
			//netwib_fmt_display("Found udp for proxy device\n"); 
            netwib_buf_append_buf(&rest, &cmpbuf);
			cmpbuf.endoffset = cmpbuf.beginoffset+8;
          	netwib_buf_cmp(&cmpbuf, &msearchbuf, &cmp);
		    if(cmp == NETWIB_CMP_EQ) {//received a M-Search
			  //check if we are proxying a device targeted by the search
			  //if so answer the search
			  netwib_er(netwib_fmt_display("M-search from ip %{ip}\n", &iphdr.src));
			  //get the service that is asked for
              get_hdr(&rest, "\r\nST:", str);
              ret = netwib_ring_index_init(proxy_devices, &pringindex);                             
              if(ret != NETWIB_ERR_OK){
                netwib_fmt_display("Error while initiating index in main.");
              }
              ret = NETWIB_ERR_OK;
              while(ret == NETWIB_ERR_OK){
                ret = netwib_ring_index_next(pringindex, &ptr);
                if(ret == NETWIB_ERR_OK){                                  
                  //found a device, send HTTP OK for service with matching ST
                  //must check all services for this device
                  upnpdev = (upnp_device *)ptr;
                  netwib_ring_index_init(upnpdev->services, &pringindex2);                                
                  while(ret == NETWIB_ERR_OK){                                
                    //get next service
                    ret = netwib_ring_index_next(pringindex2, &ptr);      
                    if(ret == NETWIB_ERR_OK){
                      //found service
                      upnp_service = (service *)ptr;
                      if(strcmp(str, "ssdp:all") == 0){                                                                   
                        //send for all services
                        b = NETWIB_TRUE;
                      }
                      else if(strcmp(str, upnp_service->st)== 0){
                        //ST matches                                                      
                        b = NETWIB_TRUE;
                      }
                      else
                        b = NETWIB_FALSE;                                                                                                                                                                 
                      if(b){                                                                                                                              
                        send_httpok(upnp_service, upnpdev, &device, &iphdr.src, &ethhdr.src, udphdr.src);                                                                                      
                      }
                    }//end of service                                                          
                  }//traversed all services for this device 
                  ret = NETWIB_ERR_OK;                                                                                                                                       
                }
                else{
                  netwib_ring_index_close(&pringindex2);                                                                    
                }                                                                        
              }//end of while
              netwib_ring_index_close(&pringindex); 
			}//end of M-search
				
			else {				
			  cmpbuf.endoffset =cmpbuf.beginoffset+ 6;
			  netwib_buf_cmp(&cmpbuf, &notifybuf, &cmp);
			  if(cmp == NETWIB_CMP_EQ) {//received a SSDP-notify
			    //update cache 
				netwib_er(netwib_fmt_display("SSDP-notify from ip %{ip}\n", &iphdr.src));
				//find out which kind of SSDP
				get_hdr(&rest, "NTS:", str);
				if(strcmp(str, "ssdp:alive") == 0){
                  //update timestamp or add to cache or remove from proxy
                  process_packet(&iphdr, &ethhdr, &udphdr, &rest, SSDP_ALIVE, upnp_devices);
                  //if device is in proxy, means we started proxying too early
                  //remove device from proxy 
                  //remove_device will first check if device is in proxy, if so it will be removed
                  ret = remove_device(&iphdr.src, proxy_devices);                                                              
                  if(ret == NETWIB_ERR_OK){
                         //must reset upnp_device.proxy to FALSE
                         find_ip(&iphdr.src, upnp_devices, &ptr);
                         upnpdev = (upnp_device *)ptr;
                         netwib_ips_del_ip(ips, &iphdr.src);
                         upnpdev->proxy = NETWIB_FALSE;
                         netwib_time_init_now(&upnpdev->last_activity);
                         netwib_fmt_display("Removed %{ip} from proxy \n",&upnpdev->ip );
                          }
                }
                else if(strcmp(str, "ssdp:byebye") == 0){                                 
                  //remove from chache and proxy
                  remove_device(&iphdr.src, proxy_devices);
                  ret = remove_device(&iphdr.src, upnp_devices);
                  if(ret == NETWIB_ERR_OK){
                         netwib_fmt_display("Removed %{ip} from cache \n",&iphdr.src);
                          }
                  
                }
			  }
			}
		  }
	    }  
		else if(ret == NETWIB_ERR_DATAOTHERTYPE) {
		  netwib_ips_contains_ip(ips, &iphdr.dst, &iptest);
		  if(iptest) {
		  	netwib_fmt_display("found tcp for proxy device\n");		 	 
		    send_wol(&ethhdr.dst); // send wol when we received a tcp message
		    //remove the device from proxy
		    ret =remove_device(&iphdr.dst, proxy_devices);
		    if(ret == NETWIB_ERR_OK){
                         //must reset upnp_device.proxy to FALSE
                         find_ip(&iphdr.dst, upnp_devices, &ptr);
                         upnpdev = (upnp_device *)ptr;
                         upnpdev->proxy = NETWIB_FALSE;
                         //update timestamp
                         netwib_time_init_now(&upnpdev->last_activity);
                         netwib_ips_del_ip(ips, &iphdr.dst);
                         netwib_fmt_display("Removed %{ip} from proxy \n",&upnpdev->ip );
                          }
		    //add forwarding of packet
		  }  
	 	}
        else if(ret == NETWIB_ERR_NOTCONVERTED){
             //there is an arp in the io
             //parse it
             ret = get_arp(piostorage, &arphdr);
             if(ret != NETWIB_ERR_OK)
                    netwib_fmt_display("Error while getting arp header");
             else{
                  netwib_ips_contains_ip(ips, &arphdr.ipdst, &iptest);
                  if(iptest){
                             //there is an arp addressed for a proxy
                             //anser it to see what is being sent
                             find_ip(&arphdr.ipdst, upnp_devices, &ptr);
                             upnpdev = (upnp_device *) ptr;
                             ret = send_arp(&arphdr.ipdst, &arphdr.ipsrc, &upnpdev->eth, &arphdr.ethsrc, 
                             &device);                             
                             netwib_fmt_display("Arp answered for %{ip} \n", &arphdr.ipdst);
                             }
                  }
             
             }	
        else if(iphdr.header.ip4.id != 123){
             //no packet found during 2 sec
        }
 
	}
	//////////////////////////////////////////////////////////////////////////////
	////////////                 close cache thread             //////////////////
	netwib_er(netwib_thread_wait(cache_check, NETWIB_TIME_INFINITE, &event, &ret, &thread_return));
	if (!event) {
      netwib_er(netwib_fmt_display("Error when closing cache update thread"));
    }
    netwib_er(netwib_thread_close(&cache_check));
    netwib_er(netwib_thread_wait(notify_check, NETWIB_TIME_INFINITE, &event, &ret, &thread_return));
	if (!event) {
      netwib_er(netwib_fmt_display("Error when closing notify check thread"));
    }
    netwib_er(netwib_thread_close(&notify_check));
  
  //////////////////////////////////////////////////////////////////////////////
	netwib_buf_close(&device);
	netwib_buf_close(&notifybuf);
	netwib_buf_close(&msearchbuf);
    netwib_er(netwib_io_close(&pio));
    netwib_er(netwib_io_close(&piostorage));
    netwib_ring_close(&upnp_devices, NETWIB_FALSE);
    netwib_ring_close(&proxy_devices, NETWIB_FALSE);
	netwib_close();
	return(0);

}


//Initiate a NIC
//input: hwt (hardware type, which kind of device is to be initiated)
//output: *pbuf (will contain description of the device), 
netwib_err init_device(netwib_buf *pbuf, netwib_device_hwtype hwt) 
{
//--------------declaration-------------------------------    
  netwib_err ret;
  netwib_conf_devices_index *pconfindex;
  netwib_conf_devices conf;
//--------------end of declaration------------------------    

  netwib_er(netwib_conf_devices_index_init(&conf, &pconfindex));
  ret = NETWIB_ERR_OK;
  while (NETWIB_TRUE) {
    ret = netwib_conf_devices_index_next(pconfindex);
    if (ret != NETWIB_ERR_OK) {
      if (ret == NETWIB_ERR_DATAEND) 
        ret = NETWIB_ERR_NOTFOUND;
      break;
    }        
    if (conf.hwtype == hwt) {
      netwib_er(netwib_buf_append_buf(&conf.device, pbuf));
      break;
    }
  }
  netwib_er(netwib_conf_devices_index_close(&pconfindex));
  if (ret == NETWIB_ERR_NOTFOUND) {
    netwib_er(netwib_fmt_display("No device of wanted type found\n"));
  }
  netwib_er(netwib_fmt_display("Device initiated\n"));
  return(ret);
}
