
/*
=============================================================
proxy_receive.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the invisible proxy program. The invisible proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover sleeping devices and make sure they still seem to be connected
to the network by spoofing HTTP OK and SSDP:alive. When there is
an incoming TCP connection for the device the proxy will wake it up using
WOL.

This file contains functions for receiving and parsing packets.
There are functions for retrieveng SSDP headers from raw packets
and a function for sniffing packets. Parsed packets are processed
and services and devices are added to the data structure.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <string.h>
#include "proxy_funcs.h"

netwib_err get_st_usn(netwib_buf *rest, char *st, char *usn );
netwib_err get_server_location(netwib_buf *rest, char *server, char *location );


//will parse the packet in *in to the headers
//input: in, dlt
//output: plinkhdr, piphdr, pudphdr, rest
netwib_err parse_packet(netwib_buf *in, netwib_device_dlttype dlt, 
                       netwib_linkhdr *plinkhdr, netwib_iphdr *piphdr,
                       netwib_udphdr *pudphdr, netwib_buf *rest) 
{ 
//--------------declaration---------------------------
      netwib_err ret;
      netwib_uint32 skipsize;
//--------------end of declaration------------------------
      
      
       ret = NETWIB_ERR_OK;
	  netwib_buf_append_buf(in, rest);
      
      ret = netwib_pkt_decode_linkhdr(dlt , rest, plinkhdr, &skipsize); // get the linkhdr
	  if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error while when decoding linkheader."));
		return(ret);
	  }
      netwib_er(netwib_buf_shift(rest, -skipsize, 0)); // shift away linkhdr
      
      if(plinkhdr->hdr.ether.type == NETWIB_ETHERHDRTYPE_ARP){
                           //we have an arp, stop decoding
                           return NETWIB_ERR_NOTCONVERTED;
                           }


      ret = netwib_pkt_decode_iphdr(rest, piphdr, &skipsize); //get the iphdr
	  if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error while when decoding ipheader."));
		return(ret);
	  }
	  
	  if(piphdr->protocol == NETWIB_IPPROTO_TCP) {
        return(NETWIB_ERR_DATAOTHERTYPE);
      }

      netwib_er(netwib_buf_shift(rest, -skipsize, 0)); // shift away iphdr

      ret = netwib_pkt_decode_udphdr(rest, pudphdr, &skipsize); //get the udphdr
      
	  if(ret != NETWIB_ERR_OK){
		netwib_er(netwib_fmt_display("Error while when decoding udpheader."));
		return(ret);
	  }
      netwib_er(netwib_buf_shift(rest, -skipsize, 0)); //shift away udphdr
      
      return(ret);                               
}





//will sniff the network 
//sniffed packet will be returned
//input: device, filter, wait (time to sniff), pio (input channel)
//output: iphdr, udphdr, ethernetaddresses and rest of the sniffed packet
netwib_err sniff(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest,
                               netwib_time *wait, netwib_io *pio)
{
//--------------declaration---------------------------
    netwib_buf packet;
	netwib_device_dlttype dlt;
	netwib_err ret;
	netwib_linkhdr linkhdr;
	netwib_bool event;
	char usn[100], st[100];

//--------------end of declaration------------------------

	ret = NETWIB_ERR_OK;
    netwib_er(netwib_buf_init_mallocdefault(&packet));    
    
//	netwib_er(netwib_fmt_display("Receiving\n")); 
	netwib_er(netwib_sniff_ctl_get_dlt(pio, &dlt)); // get dlt
			
	netwib_er(netwib_io_wait_read(pio, wait, &event)); // wait for data in io

    //does not always give an event when there is something to read
	//read from the io, get the packet
	ret = netwib_io_read(pio, &packet);

	if(ret == NETWIB_ERR_NOTCONVERTED) {
		 netwib_er(netwib_fmt_display("Error while receiving: Data not converted"));		 
		  netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if(ret == NETWIB_ERR_DATAMISSING) {
		 netwib_er(netwib_fmt_display("Error while recieving: Datamissing"));
		  netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if (ret == NETWIB_ERR_DATANOTAVAIL) {
		 netwib_er(netwib_fmt_display("."));
		 netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if (ret == NETWIB_ERR_OK) {
	//	 netwib_er(netwib_fmt_display("something received\n"));
		//parse the received packet into headers
		 ret = parse_packet(&packet, dlt, &linkhdr, iphdr, udphdr, rest);     
		 if(ret != NETWIB_ERR_OK && ret!= NETWIB_ERR_DATAOTHERTYPE ){
		//this has to be an arp
		//put data back to the io and exit
		netwib_io_unread(pio, &packet);
		return NETWIB_ERR_NOTCONVERTED;		
		        
		 }

	}

	*ethhdr = linkhdr.hdr.ether; //only need the ethernet header from the linkheader
	//close open buffers and exit	
	netwib_buf_close(&packet); 
	return(ret);
		
}

//find a header in a SSDP packet and retrieve it
//used for getting NTS, ST, USN etc from HTTP
//input: *rest (the buffer that contains the whole HTTP message), *phdr (header to locate)
//output: *pstr (must be vector)
netwib_err get_hdr(netwib_buf *rest, netwib_string hdr, char *pstr ){
//--------------declaration-------------------------------
netwib_string bufstring;
char *pc;
netwib_err ret;
int i;
//--------------end of declaration------------------------        

     ret = NETWIB_ERR_OK;
     //copy buffer to a string
     netwib_er(netwib_buf_ref_string(rest, &bufstring));
     //find ST
     pc = netwib_c_strcasestr(bufstring, hdr);    
     pc = pc+netwib_c_strlen(hdr); //skip header and :
     //skip space
     if(*pc == ' ')
            pc++;
     i = 0;
     while((pstr[i] = *pc) != '\r'){
             ++i;
             pc++;         
     }
     pstr[i] = '\0';
     
}
//find the timeout (Cache-control max-age) and parse it to an int value
//input: *rest (packet)
//output: *timeout (the int value in seconds of the timeout)
netwib_err get_timeout(netwib_buf *rest, netwib_uint16 *timeout){
//--------------declaration-------------------------------
netwib_string bufstring;
char *pc;
netwib_err ret;
netwib_uint16 tmp;
int i;
//--------------end of declaration------------------------ 
     ret = NETWIB_ERR_OK;
     //copy buffer to a string
     netwib_er(netwib_buf_ref_string(rest, &bufstring));
     //find = that is before numerical value
     pc = netwib_c_strcasestr(bufstring, "=");     
     pc++; //skip =
     //skip space
     if(*pc == ' ')
            pc++;
     tmp = 0;
     //convert string to int
     for(; *pc >='0' && *pc <='9'; pc++)
           tmp = 10*tmp+(*pc - '0');
     *timeout = tmp;
     return (ret);
      
}

//adds the packet to the ring, if it isn't already there
//can add HTTP OK or SSDP:alive
//input: *iphdr, *ethhdr, *udphdr, *rest (the packet), pt (type of packet) 
//*upnp_devices (linked list to add to)
netwib_err process_packet(netwib_iphdr *iphdr, netwib_etherhdr *ethhdr, 
							   netwib_udphdr *udphdr, netwib_buf *rest,
                               packet_type pt, netwib_ring *upnp_devices){
//--------------declaration-------------------------------
upnp_device *upnpdev;
service *upnp_service;
char str[100];
netwib_ptr ptr;
netwib_err ret;
//--------------end of declaration------------------------ 

       //search for device on IP
       ret = find_ip(&iphdr->src, upnp_devices, &ptr);                    		       
      //netwib_fmt_display("Received from ip: %{ip}\n", &iphdr->src);
   if(ret == NETWIB_ERR_OK){
       //if device is in list
       //netwib_fmt_display("device is in list\n");
       upnpdev = (upnp_device *)ptr;
       get_hdr(rest, "USN:", str);
       ret =find_usn(str, upnpdev->services, &ptr);                     
       if(ret == NETWIB_ERR_NOTFOUND){
              //and service is not in list, add the service
              netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));                              
 	          upnp_service = (service *) ptr;
              if(pt == HTTP_OK)  	         	          
 	                get_hdr(rest, "\r\nST:", upnp_service->st);
              else if (pt == SSDP_ALIVE)
                   get_hdr(rest, "\r\nNT:", upnp_service->st);
                   
              get_hdr(rest, "USN:", upnp_service->usn);
 	          get_timeout(rest, &upnp_service->timeout);	          
 	          upnp_service->port = udphdr->src;
 	          netwib_er(netwib_time_init_now(&upnp_service->last_notify));
 	          ptr = (netwib_ptr *)upnp_service;
 	          netwib_ring_add_first(upnpdev->services, ptr); 
              netwib_fmt_display("Added service to device %{ip}\n", &upnpdev->ip);      
       }
       else if(pt == SSDP_ALIVE){//if device in list and service in list
       //netwib_fmt_display("device is in list and service is in list\n");
         //and if packet is SSDP:alive, update timestamp
              upnp_service = (service *) ptr;
              netwib_er(netwib_time_init_now(&upnp_service->last_notify));
              netwib_er(netwib_time_init_now(&upnpdev->last_activity));
              netwib_fmt_display("Timestamp updated\n");
       }                               	         	
   }	         	
 	
 	else{
         //Device is not in list, add it
        //create a new device
        //netwib_fmt_display("device is not list\n");
     	netwib_er(netwib_ptr_malloc(sizeof(upnp_device), &ptr));
     	upnpdev = (upnp_device *) ptr;
     	upnpdev ->ip = iphdr->src;
     	upnpdev ->eth = ethhdr->src;
     	get_hdr(rest,"SERVER:", upnpdev ->server);
     	get_hdr(rest,"LOCATION:", upnpdev ->location);
     	upnpdev->proxy = NETWIB_FALSE;
     	netwib_er(netwib_ring_init(&delete_service_from_list, NULL, &upnpdev->services));
     	netwib_er(netwib_time_init_now(&upnpdev->last_activity));
     	ptr = (netwib_ptr *)upnpdev;
     	netwib_ring_add_first(upnp_devices, ptr);
     	//add the service to the device
     	netwib_er(netwib_ptr_malloc(sizeof(service), &ptr));
     	upnp_service = (service *) ptr;
     	if(pt == HTTP_OK)
     	      get_hdr(rest, "\r\nST:", upnp_service->st);
        else if(pt == SSDP_ALIVE)
            get_hdr(rest, "\r\nNT:", upnp_service->st);
            
     	get_hdr(rest, "USN:", upnp_service->usn);
     	get_timeout(rest, &upnp_service->timeout);
     	netwib_er(netwib_time_init_now(&upnp_service->last_notify));
     	upnp_service->port = udphdr->src;
     	ptr = (netwib_ptr *)upnp_service;
     	netwib_ring_add_first(upnpdev->services, ptr);
     	netwib_fmt_display("Added device\n");       	
           
       }                                           
}
//get an arp packet from the io
//input: *pio
//output *arp
netwib_err get_arp(netwib_io *pio, netwib_arphdr *arp){
//--------------declaration-------------------------------
netwib_err ret;
netwib_buf packet;
netwib_uint32 skipsize;
//--------------end of declaration------------------------
//there has to be an arp packet here 
        netwib_er(netwib_buf_init_mallocdefault(&packet));
    ret = netwib_io_read(pio, &packet);

	if(ret == NETWIB_ERR_NOTCONVERTED) {
		 netwib_er(netwib_fmt_display("Error while receiving: Data not converted"));		 
		  netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if(ret == NETWIB_ERR_DATAMISSING) {
		 netwib_er(netwib_fmt_display("Error while recieving: Datamissing"));
		  netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if (ret == NETWIB_ERR_DATANOTAVAIL) {
		 netwib_er(netwib_fmt_display("IO empty.\n"));
		 netwib_buf_close(&packet);
		 return(NETWIB_ERR_NOTFOUND);
	}
	else if (ret == NETWIB_ERR_OK) {
	//	 netwib_er(netwib_fmt_display("something received\n"));
		//parse the received packet into headers
		 netwib_er(netwib_buf_shift(&packet, -14, 0)); // shift away linkhdr
		 netwib_er(netwib_pkt_decode_arphdr(&packet, arp, &skipsize));
		 return NETWIB_ERR_OK;       
		 }
         
}



