/*
=============================================================
proxy_updatecache.c
Author: Martin Olsson, molsson@csee.usf.edu
and Jakob Klamra jklamra@csee.usf.edu

This file is part of the invisible proxy program. The invisible proxy is a 
proxy that allows UPnP devices to enter power sleep mode. The proxy
will discover sleeping devices and make sure they still seem to be connected
to the network by spoofing HTTP OK and SSDP:alive. When there is
an incoming TCP connection for the device the proxy will wake it up using
WOL.

This file contains functions used by threads to search the data structures.
The threads check if it is time to start proxying for a device and if
it is time to spoof SSDP:alive for a device we are proxying for.

The program uses the library netwib, which must be installed before
the program can be compiled or run. Newib can be found here:
http://www.laurentconstantin.com/en/netw/netwib/
Netwib uses Winpcap which can be found here: http://winpcap.polito.it/

This code should be compiled using Visual C++ or Dev C++. Please refer
to netwib documentation for settings of the compiler.

This program was downloaded from http://www.csee.usf.edu/~jklamra/upnp/
and is part of a master thesis about power management in Universal Plug and Play.
Please refer to the homepage for known bugs and comments.

We do not guarantee the functionality of this program.
Use this program at your own risk.
=============================================================
*/

#include <stdio.h>
#include <memory.h>
#include "proxy_funcs.h"

//Thread to move devices into proxy mode
//input: infosin a *ring containing upnp_devices ring and proxy_devices ring
//will move a device into the proxy_devices ring after x minutes of idle time
netwib_err update_cache(netwib_ptr infosin, netwib_ptr *pinfosout) 
{
//--------------declaration------------------------------- 
  netwib_ring_index *ri, *ri2;
  netwib_ptr ptr, ptr2, ptr3, ptr4;
  netwib_err ret, ret2;
  upnp_device *device, *device2;
  netwib_time t;
  netwib_cmp timecheck;
  netwib_ring *ring;
  netwib_ring *upnp_devices;
  netwib_ring *proxy_devices;
  netwib_ips *ips;
//--------------end of declaration------------------------   
  ring = (netwib_ring *)infosin;
  netwib_er(netwib_ring_index_init(ring, &ri));
  ret = netwib_ring_index_next(ri, &ptr);
  proxy_devices = (netwib_ring *)ptr;
  ret = netwib_ring_index_next(ri, &ptr);
  upnp_devices = (netwib_ring *)ptr;
  ret = netwib_ring_index_next(ri, &ptr);
  ips = (netwib_ips *)ptr;
  
  //infinite loop to update proxy_devices
  while(NETWIB_TRUE)
  {
    ret = NETWIB_ERR_OK;
    netwib_er(netwib_ring_index_init(upnp_devices, &ri));
    if(ret != NETWIB_ERR_OK){
       netwib_fmt_display("Error while initiating index in update cache.");
       return(NETWIB_ERR_NOTFOUND);
    }
    ret = netwib_ring_index_next(ri, &ptr);
    //loop to go through all devices every time
    while(ret != NETWIB_ERR_DATAEND) 
	{
      netwib_er(netwib_time_init_now(&t));
	  netwib_er(netwib_time_minus_sec(&t, 60*min_bef_proxy)); //start proxying after x*60 seconds 
	  device = (upnp_device *)ptr;
	  netwib_time_cmp(&device->last_activity, &t, &timecheck);
	  if( timecheck != NETWIB_CMP_GT && !device->proxy) {
	  	netwib_fmt_display("Found device to proxy for\n");
		device->proxy = NETWIB_TRUE;
		netwib_ips_add_ip(ips, &(device->ip)); //add device's ip to ip list
		netwib_er(netwib_ptr_malloc(sizeof(upnp_device), &ptr2));
		//copy the device to proxy_devices
		netwib_c_memcpy(ptr2, ptr, sizeof(upnp_device));
		device2 = (upnp_device *)ptr2;
		//initiate a new ring
		netwib_er(netwib_ring_init(&delete_service_from_list, NULL, &device2->services));
		netwib_er(netwib_ring_index_init(device->services, &ri2));
		ret2 = NETWIB_ERR_OK;
		//copy every item from the first ring to the second
		ret2 = netwib_ring_index_next(ri2, &ptr3);
		while(ret2 != NETWIB_ERR_DATAEND){                        
          netwib_er(netwib_ptr_malloc(sizeof(service), &ptr4));
          netwib_c_memcpy(ptr4, ptr3, sizeof(service));
          netwib_ring_add_first(device2->services, ptr4);
          ret2 = netwib_ring_index_next(ri2, &ptr3);
        }		  	 
	   	netwib_er(netwib_ring_add_first(proxy_devices, ptr2));// add device to proxy list
	  }
	  ret = netwib_ring_index_next(ri, &ptr);
    }
    netwib_er(netwib_ring_index_close(&ri));
    netwib_time_sleep_sec(5); //sleep 5 sec
  } 
}

//thread to send out SSDP:alive whenever needed
//input infosin a *ring containing upnp_devices ring and proxy_devices ring
//will send out SSDP:alive for devieces in proxymode after 1/3 of the timout time since last SSDP:alive was sent
netwib_err spoof_notify(netwib_ptr infosin, netwib_ptr *pinfosout)
{
//--------------declaration------------------------------- 
  netwib_ring_index *ri;
  netwib_ring_index *rid; 
  netwib_ring_index *ris;
  netwib_err ret1, ret2;
  netwib_time t;
  netwib_ptr ptr1;
  netwib_ptr ptr2;
  upnp_device *device;
  service *serv;
  netwib_cmp timecheck;
  netwib_ring *ring;
  netwib_ring *upnp_devices;
  netwib_ring *proxy_devices;
//--------------end of declaration------------------------ 
  ring = (netwib_ring *)infosin;
  netwib_er(netwib_ring_index_init(ring, &ri));
  ret1 = netwib_ring_index_next(ri, &ptr1);
  proxy_devices = (netwib_ring *)ptr1;
  ret1 = netwib_ring_index_next(ri, &ptr1);
  upnp_devices = (netwib_ring *)ptr1;
  //inifinite loop to send SSDP:alive 
  while(NETWIB_TRUE)
  {
    ret1 = NETWIB_ERR_OK;
    netwib_er(netwib_ring_index_init(proxy_devices, &rid));
    ret1 = netwib_ring_index_next(rid, &ptr1);
    //loop to go through all devices in proxy_devices
    while(ret1 != NETWIB_ERR_DATAEND) 
    {
      //netwib_fmt_display("found a device in proxy list\n");
	  device = (upnp_device *)ptr1;
	  ret2 = NETWIB_ERR_OK;
	  netwib_er(netwib_ring_index_init(device->services, &ris));
	  ret1 = netwib_ring_index_next(ris, &ptr2);
	  //loop to go through all services of a device
	  while(ret2 != NETWIB_ERR_DATAEND)
	  {
   	    serv = (service *) ptr2;
		netwib_er(netwib_time_init_now(&t));
		netwib_er(netwib_time_minus_sec(&t, (serv->timeout/3)));
		netwib_time_cmp(&(serv->last_notify), &t, &timecheck);
		//it is time to send a SSDP:alive
		if(timecheck == NETWIB_CMP_LT) {
          send_alive(serv, device);// send SSDP:alive for service and update last_notify
		}
		ret2 = netwib_ring_index_next(ris, &ptr2);
	  }
	  ret1 = netwib_ring_index_next(rid, &ptr1);
	}
  	netwib_time_sleep_sec(5);	//sleep for 5 seconds  
  }	   
}
